%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS save.ql,v 1.8 1993/12/10 02:23:16 marku Exp


/*
** save the current thy 
** also save table of symbols and object variables
*/

save :-
	debugpause((
		save_global_data,
		save_theory_data
		)).

/*
** Save each updated Theory
*/

save_theory_data :-
	updated_dir(Theory),
	file_name(Theory, FileName),
	save_to(FileName),
	save_thy_lines(Theory),
	saved,
	retract(updated_dir(Theory)),
	fail.

save_theory_data.


/*
** Save each instance of each fact in a given theory
*/

save_thy_lines(Theory) :-
	saved_fact(Name/Arity),
	nl,
	'$atoms_to_atom'([Name, Theory], NewHead),
	DBTerm =.. [NewHead, Params],
	clause(DBTerm, Body),
	save_thy_line(DBTerm, Body),
	fail.

save_thy_lines(Theory) :-
	([save_clauses, Theory] : [Name, Arity, ThyPos]),
	functor(Head, Name, Arity),
	(ThyPos > 0 ->
	    arg(ThyPos, Head, Theory)
	;
	    true
	),
	nl,
	clause(Head, Body),
	save_thy_line(Head, Body),
	fail.

save_thy_lines(Theory).

save_thy_line(Clause, true) :-
	!,
	writeq(Clause),
	writeln('.').

save_thy_line(Clause, Body) :-
	!,
	writeq(Clause),
	writeln(' :-'),
	save_clause_body(Body).


save_clause_body((A,B)) :-
	!,
	tab(8),writeq(A),writeln(','),
	save_clause_body(B).

save_clause_body(B) :-
	tab(8),writeq(B),writeln('.').


/*
** save global data
*/


save_global_data :-
	retract(updated_dir(global)), !,
	save_to('global.erg'),
	save_global,
	saved.

save_global_data.

save_global :-
	global_fact(Name/Arity),
        length(Params, Arity),
        nl,
        clause(Name : Params, Body),
        save_thy_line((Name : Params), Body),
        fail.



save_global.


/*
**
*/

proof_file_name(Theory:Name, FileName) :-
	atom_append(Name, '.pf', PfName),
	'$atoms_to_atom'([Theory, PfName], FileName).

file_name(Theory, FileName) :-
	atom_append(Theory, '.erg', FileName).

save_to(File) :-
	write('Saving '), write(File), write(' ... '),
	flush(user_output),
	tell(File).

saved :-
	told,
	writeln('saved.').

