%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS unix.ql,v 1.16 1993/12/10 02:23:21 marku Exp


atom_append(N1, N2, N3) :-
    atom_chars(N1, N1name),
    atom_chars(N2, N2name),
    append(N1name, N2name, N3name),
    atom_chars(N3, N3name).

/*
** Deal with the command line arguments
*/

cmdline_args([]).

cmdline_args(['-V' | Tail]) :-
	!,
	assertz(verbose_file_access_mode),
	cmdline_args(Tail).

cmdline_args(['-G', Global | Tail]) :-
	!,
	assertz(global_file_location(Global)),
	cmdline_args(Tail).

cmdline_args(['-lib', L | Tail]) :-
	!,
	assertz(library_location(L)),
	cmdline_args(Tail).

cmdline_args(['-L' | Files]) :-
	!,
	note_loaded(Files, Tail),
	cmdline_args(Tail).
	
cmdline_args(['-H' , HelpCmd | Tail]) :-
	!,
	abolish(help_cmd,1),
	assertz(help_cmd(HelpCmd)),
	cmdline_args(Tail).

cmdline_args([Arg | Tail]) :-
	writelist(['Unknown argument: ', Arg, nl]), 
	cmdline_args(Tail).


note_loaded([], []).

/* stop when we find an argument that starts with '-' */
note_loaded([Arg|Tail], [Arg|Tail]) :-
	atom_chars(Arg, [45|_]),
	!.

note_loaded([File|L], Tail) :-
	assert(loaded(File)), 
	writelist(['Pre-loaded: ', File, nl]),
	note_loaded(L, Tail).
	

/*
** Prettier consult, which understands the load path.
*/


full_file_name('.', File, '.', File) :-
	!.

full_file_name(lib(P), File, Dir, FullFile) :-
	!,
	library_location(L),
        '$atoms_to_atom'([L, P], Dir),
        '$atoms_to_atom'([L, P, File], FullFile).

full_file_name(P,   File, P, FullFile) :-
        '$atoms_to_atom'([P, File], FullFile).


do_consult(File) :-
	loaded(File), !.

do_consult(File) :-
	writelist(['Loading ', File, ' ...  ']),
	flush(user_output),
        (path : [P]),
	full_file_name(P, File, Dir, PSFile),
	(verbose_file_access_mode ->
	    writelist([nl,tab(16),'trying to read: ',PSFile,nl])
	;
	    true
	),
	consult(PSFile), !,
	writelist([tab, 'from ', Dir, nl]),
	assert(loaded(File)).

do_consult(File) :-
	writelist(['Unable to find file: ', File, nl]), nl, fail.


consult_global :-
	global_file_location(G), % set by -G arguments.
	(verbose_file_access_mode ->
	    writelist([tab(16),'trying to read: ',G,nl])
	;
	    true
	),
	consult(G),
	!,
	assert(loaded('global.erg')).

consult_global :-
	writelist(['Unable to find ''global.erg'' file.',nl,
	           'Use -G <filename> to specify a global file.',nl]),
        !,
	fail.


/*
** Start the hypertext help program
*/

help :-
	help_cmd(H),
	atom_chars(H, Hstr),
	(library_location(L) ->
	    atom_chars(L, Lstr)
	;
	    Lstr = []
	),
	append(["sh -c '", Hstr, " ", Lstr, "'"], CmdStr),
	system(CmdStr).


/*
**
*/

set_path(L) :-
	\+  (elem(E, L),
	     check_atom(E),
	     \+  (path : [E]),
	     \+  assert_global(path, [E])).
