%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS window.ql,v 1.11 1993/12/10 02:23:50 marku Exp


?- op(990,xfy,(-->)).


/*
** open window on focus in given position (returns leftover position
** information in Rest)
*/

open_pos(Path, Pos, Rest, Tactic) :-
	get_window(PT, [(F, J)|FL], FD, E, G, HL, NH),
	CW is PT + 1,

	get_up(openrule, [F,Pos,SF,NewHyps,HFn,EFn,GFn,CF-->NF]),
	match_paths(Path, Pos, Rest, Obvars),
	translate_relation(EFn, E, NE),
	!,
		
	assimilate_new_vars(Obvars),

        (member(G-->NG, GFn) ; (NG = unknown)),

	set_window(CW, [(SF, initial_focus)], FD, NE, NG),
	gen_hyps(HFn, F, HL, Obvars, NewHyps),
	next_wrap(+ WrapNr),
	!,
	
	Tactic,
	!,

	complete_window(NE, Proof, CF, CD),
	set_window(PT, [(NF,open_pos(WrapNr,Pos,Proof)), (F,J) | FL], CD, E, G),
	ipSet(hypnr, NH),
	ipSet(hyplist, HL),
        !.


translate_relation(EFn, E, NE) :-
	member(O1 --> O2, EFn),
	(
	    stronger(O1, E), 
	    equal_to(O2, NE)
	;
	    equal_to(rev(O1), RO1), 
	    stronger(RO1, E), 
	    equal_to(rev(O2), NE)
	),
	!.
	
	



match_paths(A, [], A, []).
match_paths([b(X)|Pos], [b(X)|NewPos], Rest, [X|Obvars]) :- 
	!, match_paths(Pos, NewPos, Rest, Obvars).

match_paths([2|Pos],    [b(X)|NewPos], Rest, [X|Obvars]) :- 
	!, match_paths(Pos, NewPos, Rest, Obvars).

match_paths([P|Pos],    [P|NewPos],    Rest, Obvars)   :- 
	!, match_paths(Pos, NewPos, Rest, Obvars).


/*	
** open window on a hypothesis 
*/

open_hyp(Hyp, Tactic) :- open_hyp(Hyp, Tactic, _).

open_hyp(Hyp, Tactic, Next) :-
	get_window(PT, FL, FDep, E, G, HL, NH),
	CW is PT + 1,
	delete_hyp(Hyp, (Nr, SF, _, _, Dep), HL),
	set_window(CW, [(SF, initial_focus)], Dep,  =>, false),

	inherit_for_openhyp(HL, New),
	ipSet(hyplist, New),

	ipLookup(in_hyp, InHyp),
	ipSet(in_hyp, true),
	next_wrap(+ WrapNr),
	!,

	Tactic,
	!,
	complete_window(=>, Proof, CF, CDep),
	set_window(PT, FL, FDep, E, G),
	Next is NH + 1,
	ipSet(hypnr, Next),
	ipSet(hyplist, [(Next,CF,_,open_hyp(WrapNr, Nr, Proof), CDep)|HL]),
	ipSet(in_hyp, InHyp),
        !.

inherit_for_openhyp([], []).
inherit_for_openhyp([(Nr,T,Del,_,Dep)|HL], [(Nr,T,_,inherited,Dep)|New]) :-
        var(Del), !,
        inherit_for_openhyp(HL, New).
inherit_for_openhyp([_|HL], New) :-
        inherit_for_openhyp(HL, New).


/*
** Discharge a conjecture
*/

discharge(Hyp, Tactic) :-
	get_window(PT, FL, FDep, E, G, HL, NH),

        get_hyp(Hyp, (Nr, SF, _, conjecture(WrapNr, Just), Dep), HL),
	var(Dep),
	thaw((Dep, WrapNr, Just)), 

	filter_dep(Nr, Dep, HL, New), 

	CW is PT + 1,
        set_window(CW, [(SF, initial_focus)], [], rev(=>), true),
        next_wrap(+ WrapNr),

	ipSet(hyplist, New),
        !,

        Tactic,
        !,

        complete_window(rev(=>), Just, true, Dep),
        set_window(PT, FL, FDep, E, G),
        ipSet(hypnr, NH),
        ipSet(hyplist, HL).

/*
** Remove deleted and clashing hypotheses
*/

filter_dep(_, _, [], []).
filter_dep(HNr, Dep, [(Nr,T,Del,_,HDep)|HL], [(Nr,T,_,inherited,HDep)|New]) :-
	HNr =\= Nr,
	var(Del),
	\+ \+ Dep = HDep, !,
	filter_dep(HNr, Dep, HL, New).
filter_dep(HNr, Dep, [_|HL], New) :-
	filter_dep(HNr, Dep, HL, New).

	

/*
**  Associate a portion of the proof with an arbitary term
*/

wrap(Nr, Term, Tactic) :-
	ipLookup(focuslist, [(OF,OJ)|L]),
	ipSet(focuslist, [(OF, initial_focus)]),
	ipLookup(hyplist, HL),
	next_wrap(Nr),
	!,
	Tactic,
	!,
	ipLookup(focuslist, L2),
	(L2 = [_] ->
	    ipSet(focuslist, [(OF,OJ)|L]) 
	;
	    append(L2, [_], [(F, J)|L3]),
	    ipSet(focuslist, [(F, wrap(Nr, Term, J, L3)), (OF, OJ) | L])
	),
	!.


/*
** Needs to be better than this:
*/
	
next_wrap(+ Nr) :- 
	var(Nr), !,
	ipLookup(wrap, Nr), 
	NextWrap is Nr + 1, 
	ipSet(wrap, NextWrap).
next_wrap(+ Nr) :-
	integer(Nr), !,
	ipLookup(wrap, Nr2), 
	maxw(Nr, Nr2, N),
	NextWrap is N + 1,
        ipSet(wrap, NextWrap).
next_wrap(Nr) :- nonvar(Nr), !.


maxw(A, B, B) :- A < B.
maxw(A, B, A).
	


/*
**
*/

set_window(CW, FL, FD, E, G) :-
	ipSet(cw, CW),
        ipSet(focuslist, FL),
        ipSet(focusdep, FD),
        ipSet(goal, G),
        ipSet(relation, E).

get_window(PT, FL, FD, E, G, HL, NH) :-
	ipLookup(cw, PT),
	ipLookup(focuslist, FL),
	ipLookup(focusdep, FD),
	ipLookup(relation, E),
	ipLookup(goal, G),
	ipLookup(hypnr, NH),
	ipLookup(hyplist, HL).


/*
** Gather the information for window closing.
**
** Check that there are no conjectures of this window 
** on which the focus depends, which remain undischarged.
*/


complete_window(E, (E, [(F, J)|FL], HL), F, Dep) :-
	ipLookup(focuslist, [(F, J)|FL]),
	ipLookup(focusdep, FD),
	'$collect_variables'(FD, var, [], Dep),
	ipLookup(hyplist, HL),
	is_justified_focus(Dep, HL).


is_justified_focus :-
	ipLookup(focusdep, FD),         
	'$collect_variables'(FD, var, [], Dep),
        ipLookup(hyplist, HL),
        is_justified_focus(Dep, HL).

is_justified_focus([], HL).
is_justified_focus([V|Dep], HL) :- 
       \+ (member((_,_,_,conjecture(_,_),D), HL),
            var(D),
            '$thaw'(D),
            \+ D = [V|Dep]).


/*
** gen_hyps(+HypTransFunc, +Focus, +HypList0, +Obvars, -HypList).
**
**  use the given opening rule to transform hyps (old to new)
**  Focus is here so its variables can be matched and frozen.
*/


gen_hyps(HypFn, Focus, HL, Obvars, NewHyps) :-
	inherit_hyps(HypFn, HL, HL2, NHL, Obvars),
        translate_hyps(HypFn, Focus, HL2, NHL, NNHL, Obvars),
	ipSet(hyplist, NNHL),
	add_hyps(NewHyps, 'introduced', []).


inherit_hyps(_,  [], [], [], _).

inherit_hyps(HF, [(_,_,D,_,_)|HL], HL2, New, Obv) :-
	D == deleted, !,
	inherit_hyps(HF, HL, HL2, New, Obv).

inherit_hyps(HF, [(Nr,T,_,_,Dep)|HL], HL2, [(Nr,T,_,inherited,Dep)|New], Obv) :-
	\+  \+  member(T --> T, HF),
	with_frozen_nfi_test(off, gen_nfi(Obv, T)),
	inherit_hyps(HF, HL, HL2, New, Obv).

inherit_hyps(HF, [(Nr, T, _, _, Dep)|HL], [(Nr, T, Dep)|HL2], New, Obv) :-
	inherit_hyps(HF, HL, HL2, New, Obv).
	

/*
** translate_hyps(+HypTransFunc, +Focus, +HypList0, +OrigHyps, -HypList)
**
**  Uses HypTransFunc to translate hypotheses in HypList0.
**  The resulting list of hypotheses is returned in HypList,
**  with the translated hypotheses coming first, followed by
**  all the (untranslated) hypotheses in OrigHyps.
*/

translate_hyps(_, _, [], A, A, _).
translate_hyps(Hfn, Focus, HL2, NHL, NNHL, Obv) :-
	(hfn(Hfn,Focus) -> 
	    true
	;
	    abolish(hfn,2),
	    thaw(Focus), % because asserting frozen variables is broken
	    assert(hfn(Hfn,Focus)),
	    freeze(Focus)
	),
	do_translate_hyps(HL2, Focus, NHL, NNHL, Obv).

do_translate_hyps([], _, A, A, _).
do_translate_hyps([(Nr,T,D)|HL2], F, NHL, [(Nr,NT,_,translate(Nr),D)|NNHL], Obv) :-
	hfn(HFn,F),
	member(T --> NT, HFn), !,
	with_frozen_nfi_test(off, gen_nfi(Obv, T)),
	do_translate_hyps(HL2, F, NHL, NNHL, Obv).

do_translate_hyps([_|HL2], F, NHL, NNHL, Obv) :-
	do_translate_hyps(HL2, F, NHL, NNHL, Obv).

gen_nfi([], _).
gen_nfi([A|B], T) :- 
	A not_free_in T,
	gen_nfi(B, T).

/*
**
*/

set_goal :-
	is_focus(F),
        ipSet(goal, F).

set_goal(G) :-
	check_term(G),
	ipSet(goal, G).
