%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS demo3.ql,v 1.15 1993/12/10 02:18:39 marku Exp


?- op(999, fx,  (h)).
?- op(890, xfx, (on_focus)).
?- op(830, xfy, (=>)).
?- op(810, xfy, (and)).


:- declare(rule_group/1).
:- declare(rule_group/2).
:- declare(rule_group/3).
:- declare(rule_group/4).

/*
** Mimic demo3.3 openwin/focus
*/


focus([], G, Code) :-
	split_all_hyps,
	do_simplifications(G),
	ctac(Code),
	do_simplifications(G).

focus((H, Path), G, Code) :-
	open_hyp(H, focus(Path, G, Code)),
	split_all_hyps.

focus([A|B], G, Code) :-
	open_pos([A|B], _, Rest, focus(Rest, G, Code)),
	do_simplifications(G).

focus(H, G, Code) :-
	hyp_form(H), 
	open_hyp(H, focus([], G, Code)),
	split_all_hyps.

focus(B) :- focus(B, standard, _).
focus(B, A)  :- focus(B, standard, A).
A on_focus B :- focus(B, standard, A).


discharge_hyp(Hyp) :- discharge(Hyp, ctac(_)).
discharge_hyp(Hyp, Code) :- discharge(Hyp, ctac(Code)).

/*
** Add a rule to a rule group
*/

rule_group(Theorem)             :- rule_group(standard, Theorem, _, _).
rule_group(Theorem, Dir)        :- rule_group(standard, Theorem, Dir, _).
rule_group(Group, Theorem, Dir) :- rule_group(Group, Theorem, Dir, _).

rule_group(Group, Theorem, Dir, From) :-
	decode_rule(Theorem, CD, Name),
        find_rule(CD:Name, Form, _),
	freeze(Form),
	!,
	choose_conditional(_, Form, Conds, Concl),
	choose_direction(Dir, Concl, In, Reln, Out),
        ((Reln = rev(R) ->
	    preorder(R))
	;
	    preorder(Reln)
	),
	!,
        tactic_assert(rule_group, [Group, In, Conds, CD:Name, Dir]).




/*
** Rewrite with immediately applicable rules from the 'standard' rule group
*/


do_simplifications :- 
	do_simplifications(standard).

do_simplifications(Group) :-
	is_focus(From),
%	nonvar(From),
%	(\+ atomic(From)),
	get_up(rule_group, [Group, From, true, Name, Dir]),
	trans(Name, Dir, _, []),
	!,
	do_simplifications(G).

do_simplifications(_).



/*
** Split all conjunct hypotheses
*/

split_all_hyps :- 
	ipLookup(safe_hyps, false),
	get_hyp(_, (N, A and B, _, J, D)),
	\+ (var(D), functor(J, conjecture, _)), !,
	split_hyp(N),  
	split_all_hyps.
split_all_hyps.
	
	

/*
** Change focus
*/

set_focus(G) :-
	trans_goal(r:falseimp, G, [h false]).



/*
** Discharge 
*/

make_discharge(true, not_used) :- !.
make_discharge(Term, trans_true(r:refleximp, List)) :-
        satisfy(Term, List).



discharge :-
        is_focus(F),
        make_discharge(F, T),
        once(T).


%%
%%  Special case optimization:
%%


discharge_focus((Hyp, [1])) :-
        is_hyp(Hyp, Num, A => _), !,
        satisfy(A, AList),
        h_trans(Num, r:mp, AList).

discharge_focus(X) :-
        discharge on_focus X.


