%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS int.ql,v 1.9 1993/12/10 02:18:45 marku Exp


:- heuristic(int, 2, 
	['Uses the compiled rule ''int''to discharge',nl,
	 'int(_) typing conditions'],
	 is_focus(int(I)),
	 'Trying compiled rule: ''int''...',
	 trans_true(int),
	 ['int rule was successful']
	 ).

:- heuristic(eval_le, 2,
	['Evaluates conditions of the form A =< B,',nl,
	 'where A and B are both literal integer expressions'],
	 (is_focus(A =< B), ((A =< B) ; (A > B))),
	['Evaluating ', A =< B],
	((A =< B ->
	    trans_true(eval_le))
	;
	    trans_dir(eval_not_le, f)
	),
	[]
	).

:- heuristic(eval_lt, 2,
	['Evaluates conditions of the form A < B,',nl,
	 'where A and B are both literal integer expressions'],
	 (is_focus(A < B), ((A < B) ; (A >= B))),
	['Evaluating ', A < B],
	((A < B ->
	    trans_true(eval_lt))
	;
	    trans_dir(eval_not_lt, f)
	),
	[]
	).

:- envir(arith, 2,
	[ [ quit, use_hyp ],
	  [ conjuncts, int, eval_le ],
	  [ localness, split_hyp ],
	  [ always(use(_, goal)) ]
	]).


:- default_envir(arith).



%%
%% int_induction(+MVar, ?ObjVar, [?Localness,?BaseCase,?SuccCase,?PredCase])
%% int_induction(+MVar, ?ObjVar, [?Localness,?BaseCase,?SuccCase])
%%

:- tactic(int_induction(M, X, _),
        ['int_induction(+Term,?ObjVar,[?Localness,?Base,?Succ,?Pred]).'],
	['Integer Induction (succ/pred step cases) on ', M],
        [['Localness Condition'], 
	 ['Base Case'], 
	 ['Step Case - succ(', X, ')'],
	 ['Step Case - pred(', X, ')']]
        ).

:- tactic(nat_induction(M, X, _),
        ['nat_induction(+Term,?ObjVar,[?Localness,?Base,?Succ]).'],
	['Natural number induction on ', M],
        [['Localness Condition'], 
	 ['Base Case'], 
	 ['Step Case - succ(', X, ')']
        ]
        ).

int_induction(MVar, ObjVar, [L, B, SC, PC]) :-
        check_or_writelist(
                induction_focus(int(MVar), SVar, Path),
                ['Focus not in the correct form for integer induction',nl]),
        init_code( SCode, SC,
            focus(Path, (forwards(succ_case, 
				  not succ(A) = 0 and 
				  not (succ(A) < 0) and 
			          int(succ(A))), SC))),
        init_code( PCode, PC,
            focus(Path, (forwards(pred_case,
				  not pred(A) = 0 and 
				  pred(A) < 0 and 
			          int(pred(A))), PC))),
	number_of_hyps(N),
	maplist([L, B, SCode, PCode], lambda x (delete_new_hyps(N), x), Code),
        generalize(MVar, ObjVar),
	!,
	trans_dir(arith:int_induction, r, []),
	!,
	conjuncts(Code).

nat_induction(MVar, ObjVar, [L, B, SC]) :-
        check_or_writelist(
	    (
		(induction_focus(MVar:nats, SVar, Path),
		 Thm = arith:nat_induction2)
	    ;
	        (induction_focus(int(MVar) and 0 =< MVar, SVar, Path),
		 Thm = arith:nat_induction)
	    ),
            ['Focus not in a correct form for natural number induction.',nl]),
        init_code( SCode, SC,
            focus(Path, (fwd(arith:succ_case, 
				  not succ(A) = 0 and 
				  not (succ(A) < 0) and 
			          int(succ(A))), SC))),
	number_of_hyps(N),
	maplist([L, B, SCode], lambda x (delete_new_hyps(N), x), Code),
        generalize(MVar, ObjVar),
	!,
	trans_dir(Thm, r, []),
	!,
	conjuncts(Code).
		


