%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS prop_auto.ql,v 1.6 1993/12/10 02:18:59 marku Exp


/*
** A heuristic to do the obvious things to propositional foci 
*/

:- heuristic(prop_focus, 2, 'Do the obvious search with propositional foci', 
	                 when_focus(_ <=> _,  eqv_cases)).
:- heuristic(prop_focus, when_focus(_ => _,   focus([2], _))).
:- heuristic(prop_focus, when_focus(not _,    focus([1], _))).
:- heuristic(prop_focus, when((is_focus(A or B), is_goal(G)), 
			      or_focus(A, B, G))).
	

or_focus(A, _, true) :-
	satisfy(A, L), !,
	open_pos([1], trans_true(r:refleximp, L)), 
	trans_true(r:true_or).

or_focus(_, B, true) :-
	satisfy(B, L), !,
	open_pos([2], trans_true(r:refleximp, L)), 
	trans_true(r:or_true).

or_focus(_, _, _) :-
	focus([1], _).

/*
** Simplify hypotheses relative to one another
*/

:- heuristic(hyp_simp, 2,
       'Looks for hypotheses that match or negate subterms of other hypotheses',
        (is_hyp(N1, H1),
         is_hyp(N2, H2),
         N1 \== N2,
         ex_path(H2, H1, Path)),
        ['Using hypothesis ',N1,' to simplify hypothesis ',N2],
        focus((N2, Path), trans_true(N1)),
        []).

:- heuristic(hyp_simp, 2,
	sub_heuristic,
	(is_hyp(N1, not H1),
	 is_hyp(N2, H2),
	 N1 \== N2,
	 ex_path(H2, H1, Path)),
        ['Using hypothesis ',N1,' to simplify (by negation) hypothesis ',N2],
        focus((N2, Path), trans(r:notimpbool, [N1])),
	[]).

%%
%% Case analysis of hypotheses
%%


:-  tactic(hyp_cases(Hyp,_),
         ['hyp_cases(Hyp) -- case analysis with hypothesis Hyp'],
         ['Case analysis'],
         ['Case  1', 'Case 2']
         ).

		 
hyp_cases(H, [C1, C2]) :-
        is_hyp(H, N, A or B),
        trans(caseor, [N]),
        delete_hyp(N),
	number_of_hyps(NH),
        conjuncts([focus([2], (delete_new_hyps(NH), C1)),
                   focus([2], C2)]).

hyp_cases(H, [C1, C2]) :-
        is_hyp(H, N, A => B),
        trans(caseimp, [N]),
        delete_hyp(N),
	number_of_hyps(NH),
        conjuncts([focus([2], (delete_new_hyps(NH), C1)),
                   focus([2], C2)]).

hyp_cases(H, L) :-
        is_hyp(H, N, not(A and B)),
        open_hyp(N, trans(r:deMorganand)),
        hyp_cases(h(not(A) or not(B)), L).

hyp_cases(H, L) :-
        is_hyp(H, N, not(A <=> B)),
        open_hyp(N, (open_pos([1], trans(r:def(<=>))),
                     trans(r:deMorganand))),
        hyp_cases(h(not(A => B) or not(B => A)), L).




/* The super-heuristic for propositional logic */

:- heuristic(prop, 1, 'Tautology checker for propositional logic',
		   use_hyp).
:- heuristic(prop, conjuncts).
:- heuristic(prop, 1, sub_heuristic,
        (is_focus(not(A)), is_goal(true)),
        ['Focus = not(A) ; push A into hyps and search for conflicts'],
        ( trans_r(impfalse), focus([2],_) ),
        []).

%%  :- heuristic(prop, always(rewrite(PropGroup))).
:- heuristic(prop, when_focus(true, quit)).
:- heuristic(prop, split_hyp).
:- heuristic(prop, hyp_simp).
:- heuristic(prop, prop_focus).
:- heuristic(prop, always(hyp_cases(_))).
:- heuristic(prop, always((is_hyp(H, N, A <=> B),
                           h_trans(N, r:def(<=>)),
			   split_all_hyps,
			   hyp_cases(h(A => B))))).

		 
/*                    ... for predicate logic */

:- heuristic(pred, 2, 'Automated prover for a subset of predicate logic',
		   prop).
:- heuristic(pred, when_focus(all x A, focus([2]))).
:- heuristic(pred, when( is_hyp(N, ex x A), skolemize(N))).
:- heuristic(pred, when_focus(ex x A, inst_ex)).
:- heuristic(pred, when( is_hyp(N, all x A), safe(h_trans(N, r:install)))).






